import pytest
from pandas.testing import assert_series_equal
from siuba.dply import vector as v
from siuba.siu import _, Call, Symbolic, strip_symbolic
from pandas import Series

# TODO: ideally all vector functions should dispatch, and return a like-type result for...
#         - numpy arrays
#         - pandas Series
#         - array likes (e.g. lists, where valid?)

@pytest.fixture
def x():
    return Series([0,1,2])

@pytest.mark.parametrize('v_func, res', [
    (v.cumall, [False, False, False]),
    (v.cumany, [False, True, True]),
    (v.cummean, [0, .5, 1]),
    (v.desc, [2,1,0]),
    (v.dense_rank, Series([1,2,3], dtype = float)),     # Note: current pandas impl returns float, not int?
    #(v.percent_rank, ["TODO"]),  # NotImplementedError
    (v.min_rank, Series([1,2,3], dtype = float)),
    (v.cume_dist, [1/3., 2/3., 3/3.]),
    (v.row_number, [1,2,3]),
    (v.lead, [1, 2, None]),
    (v.lag, [None, 0, 1]),
    # TODO: v.near, v.nth, v.first, v.last
    ])
def test_vector_unary_and_dispatch(x, v_func, res):
    target = Series(res) if not isinstance(res, Series) else res

    assert_series_equal(v_func(x), target)

    # symbolic call argument
    sym_call = v_func(_)
    assert isinstance(sym_call, Symbolic)
    assert_series_equal(sym_call(x), target)

    # call argument
    call = v_func(strip_symbolic(_))
    assert isinstance(call, Call)
    assert_series_equal(call(x), target)


# n needs own test, since returns 0 dim object (int)
# this allows broadcasting in cases like some_col + n(some_col)
def test_vector_n(x):
    assert v.n(x) == 3

def test_vector_n_distinct(x):
    assert v.n_distinct(x) == 3

# Non-unary vector funcs ----

def test_vector_between(x):
    out = v.between(x, 0, 1)
    assert_series_equal(out, Series([True, True, False]))

def test_vector_na_if(x):
    out = v.na_if(x, [0, 2])
    assert_series_equal(out, Series([None, 1, None]))


# Need to be implemented ----

@pytest.mark.skip("TODO: Not Implemented")
def test_vector_percent_rank(x):
    pass

@pytest.mark.skip("TODO: Not Implemented")
def test_vector_coalesce(x):
    pass

@pytest.mark.skip("TODO: Not Implemented")
def test_vector_ntile(x):
    pass

@pytest.mark.skip("TODO: Not Implemented")
def test_vector_near(x):
    pass

@pytest.mark.skip("TODO: Not Implemented")
def test_vector_nth(x):
    pass

@pytest.mark.skip("TODO: Not Implemented")
def test_vector_first(x):
    pass

@pytest.mark.skip("TODO: Not Implemented")
def test_vector_last(x):
    pass



import itertools

# TODO: symbolic formatting: __add__ -> "+"

BINARY_LEVELS = {
        "__add__": 2,
        "__sub__": 2,

        "__mul__": 1,
        "__matmul__": 1,
        "__truediv__": 1,
        "__floordiv__": 1,
        "__mod__": 1,
        "__divmod__": 1,

        "__pow__": 0,

        "__lshift__": 3,
        "__rshift__": 3,

        "__and__": 4,
        "__xor__": 4,
        "__or__": 4,
        "__gt__": 5,
        "__lt__": 5,
        "__eq__": 5,
        "__ne__": 5,
        "__ge__": 5,
        "__le__": "5",

        "__getattr__": 0
        }

BINARY_OPS = {
        "__add__": "+",
        "__sub__": "-",
        "__mul__": "*",
        "__matmul__": "@",
        "__truediv__": "/",
        "__floordiv__": "//",
        "__mod__": "%",
        "__divmod__": "IDK",
        "__pow__": "**",
        "__lshift__": "<<",
        "__rshift__": ">>",
        "__and__": "&",
        "__xor__": "^",
        "__or__": "|",
        "__gt__": ">",
        "__lt__": "<",
        "__eq__": "==",
        "__ne__": "!=",
        "__ge__": ">=",
        "__le__": "<=",
        "__getattr__": "."
        }

UNARY_OPS = {
        "__invert__": "~",
        "__neg__": "-",
        "__pos__": "+",
        "__abs__": "ABS => "
        }

# TODO: can it just be put in binary ops? Special handling in Symbolic class?
MISC_OPS = {
        "__getitem__": "["
        }

ALL_OPS = {**BINARY_OPS, **UNARY_OPS, **MISC_OPS}

for k, v in BINARY_OPS.copy().items():
    BINARY_OPS[k.replace("__", "__r", 1)] = v

for k, v in BINARY_LEVELS.copy().items():
    BINARY_LEVELS[k.replace("__", "__r", 1)] = v

class Formatter:
    def __init__(self): pass

    def format(self, call):
        fmt_block = "█─"
        fmt_pipe = "├─"
        
        # TODO: why are some nodes still symbolic?
        if isinstance(call, Symbolic):
            return self.format(call.source)

        if isinstance(call, MetaArg):
            return "_"

        if isinstance(call, Call):
            call_str = fmt_block + ALL_OPS.get(call.func, repr(call.func))

            args_str = (self.format(arg) for arg in call.args)

            # TODO: kwargs handling looks funny.. (e.g. _.a(b = _.c))
            kwargs_str = (k + " = " + self.format(v) for k,v in call.kwargs.items())

            all_args = [*args_str, *kwargs_str]
            if len(all_args):
                fmt_args = [*map(self.fmt_pipe, all_args[:-1]), self.fmt_pipe(all_args[-1], final = True)]
            else:
                fmt_args = []
            return "".join([call_str, *fmt_args])

        call_str = repr(call)
        indx = call_str.find("\n")

        if indx != -1:
            return call_str
        else:
            return call_str

    @staticmethod
    def fmt_pipe(x, final = False):
        if not final:
            connector = "\n│ " if not final else "\n  "
            prefix = "\n├─"
        else:
            connector = "\n  "
            prefix = "\n└─"
        return prefix + connector.join(x.splitlines())


# Calls
# =============================================================================

class Call:
    """
    Representation of python operations.

    This class is responsible for representing the pieces of a python expression,
    as a function, along with its args and kwargs.

    For example, "some_object.a" would be represented at the function "__getattr__",
    with the args `some_object`, and `"a"`.

    Args:
        func: name of the function called. Class methods are represented using the names
              they have when defined on the class.
        *args: arguments the func call uses.
        **kwargs: keyword arguments the func call uses.


    """
    def __init__(self, func, *args, **kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def __repr__(self):
        # TODO: format binary, unary, call, associative
        if self.func in BINARY_OPS:
            op_repr = BINARY_OPS[self.func]
            fmt = "({args[0]} {func} {args[1]})"
        elif self.func in UNARY_OPS:
            op_repr = UNARY_OPS[self.func]
            fmt = "({func}{args[0]})"
        elif self.func == "getattr":
            op_repr = "."
            fmt = "({args[0]}{func}{args[1]})"
        else:
            op_repr, *arg_str = map(repr, self.args)
            kwarg_str = (str(k) + " = " + repr(v) for k,v in self.kwargs.items())

            combined_arg_str = ",".join(itertools.chain(arg_str, kwarg_str))
            fmt = "{}({})".format(op_repr, combined_arg_str)
            return fmt

        return fmt.format(
                    func = op_repr or self.func,
                    args = self.args,
                    kwargs = self.kwargs
                    )

    def __call__(self, x):
        inst, *rest = (self.evaluate_calls(arg, x) for arg in self.args)
        kwargs = {k: self.evaluate_calls(v, x) for k, v in self.kwargs.items()}
        
        # TODO: temporary workaround, for when only __get_attribute__ is defined
        if self.func == "__getattr__":
            return getattr(inst, *rest)

        # in normal case, get method to call, and then call it
        return getattr(inst, self.func)(*rest, **kwargs)

    @staticmethod
    def evaluate_calls(arg, x):
        if isinstance(arg, Call): return arg(x)

        return arg

    def copy(self):
        args, kwargs = self.map_subcalls(self.copy)
        return self.__class__(self.func, *args, **kwargs)

    def map_subcalls(self, f):
        new_args = tuple(f(arg) if isinstance(arg, Call) else arg for arg in self.args)
        new_kwargs = {k: f(v) if isinstance(v, Call) else v for k,v in self.kwargs.items()}

        return new_args, new_kwargs

    def iter_subcalls(self, f):
        yield from iter(arg for arg in self.args if instance(arg, Call))
        yield from iter(v for k,v in self.kwargs.items() if isinstance(v, Call))

    def op_vars(self, attr_calls = True):
        """Return set of all variable names used in Call
        
        Args:
            attr_calls: whether to include called attributes (e.g. 'a' from _.a())
        """
        varnames = set()

        op_var = self._get_op_var()
        if op_var is not None:
            varnames.add(op_var)

        if (not attr_calls
            and self.func == "__call__"
            and isinstance(self.args[0], Call)
            and self.args[0].func == "__getattr__"
            ):
            # skip obj, since it fetches an attribute this node is calling
            prev_obj, prev_attr = self.args[0].args
            all_args = itertools.chain([prev_obj], self.args[1:], self.kwargs.values())
        else:
            all_args = itertools.chain(self.args, self.kwargs.values())


        for arg in all_args:
            if isinstance(arg, Call):
                varnames.update(arg.op_vars(attr_calls = attr_calls))

        return varnames
    
    def _get_op_var(self):
        if self.func in ("__getattr__", "__getitem__") and isinstance(self.args[1], str):
            return self.args[1]

    def obj_name(self):
        obj = self.args[0]
        if isinstance(obj, Call):
            if obj.func == "__getattr__":
                return obj.args[0]
        elif hasattr(obj, '__name__'):
            return obj.__name__

        return None


class Lazy(Call):
    def __init__(self, func, arg = None):
        if arg is None:
            self.func = "<lazy>"
            self.args = [func]
        else:
            # will happen in generic node calls, e.g. self.__class__(self.func, ...)
            self.func = func
            self.args = [arg]

        self.kwargs = {}

    def __call__(self, x, *args, **kwargs):
        return self.args[0]



class UnaryOp(Call):
    def __repr__(self):
        fmt = "{func}{args[0]}"

        func = UNARY_OPS[self.func]
        return fmt.format(func = func, args = self.args, kwargs = self.kwargs)


class BinaryOp(Call):
    def __repr__(self):
        level = BINARY_LEVELS[self.func]
        spaces = "" if level == 0 else " "
        #if level < 4:
        #    spaces = " "*level
        #    fmt = "{args[0]}{spaces}{func}{spaces}{args[1]}"
        #else:
        #    spaces = " "
        #    fmt = "({args[0]}{spaces}{func}{spaces}{args[1]})"
        args = self.args
        arg0 = "({args[0]})" if self.needs_paren(args[0]) else "{args[0]}"
        arg1 = "({args[1]})" if self.needs_paren(args[1]) else "{args[1]}"
        fmt = arg0 + "{spaces}{func}{spaces}" + arg1


        func = BINARY_OPS[self.func]
        return fmt.format(func = func, args = self.args, kwargs = self.kwargs, spaces = spaces)

    def needs_paren(self, x):
        if isinstance(x, BinaryOp):
            sub_lvl = BINARY_LEVELS[x.func]
            level = BINARY_LEVELS[self.func]
            if sub_lvl != 0 and sub_lvl != level:
                return True

        return False


class DictCall(Call):
    """evaluates both keys and vals."""

    def __init__(self, f, *args, **kwargs):
        # TODO: validation, clean up class
        super().__init__(f, *args, **kwargs)
        self.args = (dict, dict(self.args[1]))

    def map_subcalls(self, f):
        d = self.args[1]

        new_d = {
                f(k) if isinstance(k, Call) else k: f(v) if isinstance(v, Call) else v
                        for k,v in d.items()
                        }

        return (self.args[0], new_d), {}

    def __call__(self, x):
        return self.args[1]

# Special kinds of call arguments ----
# These functions insure that when using siu expressions generated by _,
# that call.args[0] is always another call. This allows them to trivially
# respond to calling a siu expression, map_subcalls, etc..
#
# In the future, could make a parent class for Call, with a restricted
# set of behavior similar to theirs.
#
# TODO: validate that call.args[0] is a Call in tree visitors?

class MetaArg(Call):
    def __init__(self, func, *args, **kwargs):
        self.func = "_"
        self.args = tuple()
        self.kwargs = {}

    def __repr__(self):
        return self.func

    def __call__(self, x):
        return x

class FuncArg(Call):
    def __init__(self, func, *args, **kwargs):
        self.func = '__custom_func__'

        if func == '__custom_func__':
            func = args[0]

        self.args = tuple([func])
        self.kwargs = {}

    def __repr__(self):
        return repr(self.args[0])

    def __call__(self, x):
        return self.args[0]


# Trees and Visitors ==========================================================
from .error import ShortException

class FunctionLookupError(ShortException): pass


class CallVisitor:
    """
    A node visitor base class that walks the call tree and calls a
    visitor function for every node found.  This function may return a value
    which is forwarded by the `visit` method.

    Note: essentially a copy of ast.NodeVisitor
    """

    def visit(self, node):
        """Visit a node."""
        method = 'visit_' + node.func
        visitor = getattr(self, method, self.generic_visit)
        return visitor(node)

    def generic_visit(self, node):
        """Called if no explicit visitor function exists for a node."""
        
        node.map_subcalls(self.visit)

    @classmethod
    def quick_visitor(cls, visit_dict, node):
        """Class method to quickly define and run a custom visitor."""

        qv = type('QuickVisitor', cls, visit_dict)
        qv.visit(node)

class CallListener:
    """Generic listener. Each exit is called on a node's copy."""
    def enter(self, node):
        method = 'enter_' + node.func
        f_enter = getattr(self, method, self.generic_enter)

        return f_enter(node)

    def exit(self, node):
        method = 'exit_' + node.func
        f_exit = getattr(self, method, self.generic_exit)
        return f_exit(node)

    def generic_enter(self, node):
        args, kwargs = node.map_subcalls(self.enter)

        return self.exit(node.__class__(node.func, *args, **kwargs))

    def generic_exit(self, node):
        return node

    def enter_if_call(self, x):
        if isinstance(x, Call):
            return self.enter(x)

        return x

def get_attr_chain(node, max_n):
    # TODO: need to make custom calls their own Call class, then will not have to
    #       do these kinds of checks, since a __call__ will always be on a Call obj
    if not isinstance(node, Call):
        return [], node

    out = []
    ttl_n = 0
    crnt_node = node
    while ttl_n < max_n:
        if crnt_node.func != "__getattr__": break

        obj, attr = crnt_node.args
        out.append(attr)

        ttl_n += 1
        crnt_node = obj

    return list(reversed(out)), crnt_node


from inspect import isclass, isfunction
from typing import get_type_hints
from .utils import is_dispatch_func_subtype

class CallTreeLocal(CallListener):
    def __init__(
            self,
            local,
            call_sub_attr = None,
            chain_sub_attr = False,
            dispatch_cls = None,
            result_cls = None
            ):
        """
        Arguments:
            local: a dictionary mapping func_name: func, used to replace call expressions.
            call_sub_attr: a set of attributes signaling any subattributes are property
                           methods. Eg. {'dt'} to signify in _.dt.year, year is a property call.
            chain_sub_attr: whether to included the attributes in the above argument, when looking up
                           up a replacement for the property call. E.g. does local have a 'dt.year' entry.
            dispatch_cls: if custom calls are dispatchers, dispatch on this class. If none, use their name
                          to try and get their corresponding local function.
            result_cls: if custom calls are dispatchers, require their result annotation to be a subclass
                          of this class.
        """
        self.local = local
        self.call_sub_attr = set(call_sub_attr or [])
        self.chain_sub_attr = chain_sub_attr
        self.dispatch_cls = dispatch_cls
        self.result_cls = result_cls

    def create_local_call(self, name, prev_obj, cls, func_args = None, func_kwargs = None):
        # need call attr name (arg[0].args[1]) 
        # need call arg and kwargs
        func_args = tuple() if func_args is None else func_args
        func_kwargs = {} if func_kwargs is None else func_kwargs

        try:
            local_func = self.local[name]
        except KeyError as err:
            raise FunctionLookupError("Missing translation for function call: %s"% name)

        if isclass(local_func) and issubclass(local_func, Exception):
            raise local_func

        return cls(
                "__call__",
                local_func,
                prev_obj,
                *func_args,
                **func_kwargs
                )

    def enter(self, node):
        # if no enter metthod for operators, like __invert__, try to get from local
        # TODO: want to only do this if func is the name of an infix op's method
        method = 'enter_' + node.func
        if not hasattr(self, method) and node.func in self.local:
            args, kwargs = node.map_subcalls(self.enter)
            return self.create_local_call(node.func, args[0], Call, args[1:], kwargs)

        return super().enter(node)

    def enter___getattr__(self, node):
        obj, attr = node.args
        if obj.func == "__getattr__" and obj.args[1] in self.call_sub_attr:
            prev_obj, prev_attr = obj.args

            # TODO: should always call exit?
            if self.chain_sub_attr:
                # use chained attribute to look up local function instead
                # e.g. dt.round, rather than round
                attr = prev_attr + '.' + attr
            return self.create_local_call(attr, prev_obj, Call)

        return self.generic_enter(node)

    def enter___custom_func__(self, node):
        func = node(None)
        
        # TODO: not robust at all, need class for singledispatch? unique attr flag?
        if (hasattr(func, 'registry')
            and hasattr(func, 'dispatch')
            and self.dispatch_cls is not None
            ):
            # allow custom functions that dispatch on dispatch_cls
            f_for_cls = func.dispatch(self.dispatch_cls)
            if (self.result_cls is None
                or is_dispatch_func_subtype(f_for_cls, self.dispatch_cls, self.result_cls)
                ):
                # matches return annotation type (or not required)
                return node.__class__(f_for_cls)
            
            raise FunctionLookupError(
                    "External function {name} can dispatch on the class {dispatch_cls}, but "
                    "must also have result annotation of (sub)type {result_cls}"
                        .format(
                            name = func.__name__,
                            dispatch_cls = self.dispatch_cls,
                            result_cls = self.result_cls
                            )
                    )

        # doesn't raise an error so we can look in locals for now
        # TODO: remove behavior, once all SQL dispatch funcs moved from locals
        return self.generic_enter(node)

    def enter___call__(self, node):
        """
        Overview:
             variables      _.x.method(1)         row_number(_.x, 1)
             ---------      -------------        --------------------
            
                            █─'__call__'         █─'__call__'                           
             obj            ├─█─.                ├─<function row_number
                            │ ├─█─.              ├─█─.                                  
                            │ │ ├─_              │ ├─_                                  
                            │ │ └─'x'            │ └─'x'                                
                            │ └─'method'         │                                   
                            └─1                  └─1
        """
        obj, *rest = node.args
        args = tuple(self.enter_if_call(child) for child in rest)
        kwargs = {k: self.enter_if_call(child) for k, child in node.kwargs.items()}

        attr_chain, target = get_attr_chain(obj, max_n = 2)
        if attr_chain:
            # want _.x.method() -> method(_.x), need to transform
            if attr_chain[0] in self.call_sub_attr:
                # e.g. _.dt.round()
                call_name = ".".join(attr_chain) if self.chain_sub_attr else attr_chain[-1]
                entered_target = self.enter_if_call(target)
            else:
                call_name = attr_chain[-1]
                entered_target = self.enter_if_call(obj.args[0])

        elif isinstance(obj, FuncArg) and self.dispatch_cls is None:
            # want function(_.x) -> new_function(_.x), has form
            call_name = obj.obj_name()
            # the first argument is basically "self"
            entered_target, *args = args
        else:
            # default to generic enter
            return self.generic_enter(node)

        return self.create_local_call(
                call_name, entered_target, node.__class__,
                args, kwargs
                )


# Also need NodeTransformer


# Symbolic
# =============================================================================

class Symbolic(object):
    def __init__(self, source = None, ready_to_call = False):
        self.source = MetaArg("_") if source is None else source
        self.ready_to_call = ready_to_call

    def __getattr__(self, x):
        # temporary hack working around ipython pretty.py printing
        #if x == "__class__": return Symbolic

        return Symbolic(BinaryOp(
                "__getattr__",
                self.source,
                strip_symbolic(x)
                ))
                

    def __call__(self, *args, **kwargs):
        if self.ready_to_call:
            return self.source(*args, **kwargs)

        return create_sym_call(self.source, *args, **kwargs)

    def __getitem__(self, *args):
        return Symbolic(Call(
                "__getitem__",
                self.source,
                *map(slice_to_call, args)
                ),
                ready_to_call = True)

    
    def __invert__(self):
        if isinstance(self.source, Call) and self.source.func == "__invert__":
            return self.source.args[0]
        else: 
            return self.__op_invert()


    def __op_invert(self):
        return Symbolic(UnaryOp('__invert__', self.source), ready_to_call = True)

        
    def __repr__(self):
        return Formatter().format(self.source)


def create_sym_call(source, *args, **kwargs):
    return Symbolic(Call(
            "__call__",
            strip_symbolic(source),
            *map(strip_symbolic, args),
            **{k: strip_symbolic(v) for k,v in kwargs.items()}
            ),
            ready_to_call = True)


def slice_to_call(x):
    if isinstance(x, slice):
        args = map(strip_symbolic, (x.start, x.stop, x.step))
        return Call("__call__", slice, *args)
    
    return strip_symbolic(x)


def str_to_getitem_call(x):
    return Call("__getitem__", MetaArg("_"), x)

    
def strip_symbolic(x):
    if isinstance(x, Symbolic):
        return x.source

    return x


def explain(symbol):
    """Print representation that resembles code used to create symbol."""
    if isinstance(symbol, Symbolic):
        print(symbol.source)
    else: 
        print(symbol)


# symbolic dispatch wrapper ---------------------------------------------------

from functools import singledispatch, update_wrapper
import inspect

def _dispatch_not_impl(func_name):
    def f(x, *args, **kwargs):
        raise TypeError("singledispatch function {func_name} not implemented for type {type}"
                            .format(func_name = func_name, type = type(x))
                            )

    return f

def symbolic_dispatch(f = None, cls = object):
    if f is None:
        return lambda f: symbolic_dispatch(f, cls)

    # TODO: don't use singledispatch if it has already been done
    dispatch_func = singledispatch(f)

    if cls is not object:
        dispatch_func.register(cls, f)
        dispatch_func.register(object, _dispatch_not_impl(dispatch_func.__name__))


    @dispatch_func.register(Symbolic)
    def _dispatch_symbol(__data, *args, **kwargs):
        return create_sym_call(FuncArg(dispatch_func), __data.source, *args, **kwargs)

    @dispatch_func.register(Call)
    def _dispatch_call(__data, *args, **kwargs):
        # TODO: want to just create call, for now use hack of creating a symbolic
        #       call and getting the source off of it...
        return create_sym_call(FuncArg(dispatch_func), __data, *args, **kwargs).source

    return dispatch_func

    
# Do some gnarly method setting -----------------------------------------------

def create_binary_op(op_name):
    def _binary_op(self, x):
        node = BinaryOp(op_name, self.source, strip_symbolic(x))

        return Symbolic(node, ready_to_call = True)

    return _binary_op

def create_unary_op(op_name):
    def _unary_op(self):
        node = UnaryOp(op_name, self.source)

        return Symbolic(node, ready_to_call = True)

    return _unary_op

for k, v in BINARY_OPS.items():
    if k in {"__getattr__"}: continue
    rop = k.replace("__", "__r")
    setattr(Symbolic, k, create_binary_op(k))

for k, v in UNARY_OPS.items():
    if k != "__invert__":
        setattr(Symbolic, k, create_unary_op(k))

Lam = Lazy

_ = Symbolic()




from siuba.siu import Symbolic, strip_symbolic
from siuba.spec.series import spec
from .helpers import data_frame, assert_equal_query, backend_pandas, SqlBackend, PandasBackend
import pytest
# TODO: dot, corr, cov

from siuba import filter, mutate, summarize, group_by
from pandas.testing import assert_frame_equal, assert_series_equal
import numpy as np
import pandas as pd
import pkg_resources

def filter_on_result(spec, types):
    return [k for k,v in spec.items() if v['result']['type'] in types]

SPEC_IMPLEMENTED = filter_on_result(spec, {"Agg", "Elwise", "Window"})
SPEC_NOTIMPLEMENTED = filter_on_result(spec, {"Singleton"})
SPEC_AGG = filter_on_result(spec, {"Agg"})

_ = Symbolic()

@pytest.fixture(params = tuple(SPEC_IMPLEMENTED))
def entry(request):
    # NOTE: the sole purpose of putting in a fixture is so pytest line output
    #       is very easy to read. (e.g. pytest -v --tb=line)
    key = request.param
    yield spec[key]

@pytest.fixture(params = tuple(SPEC_AGG))
def agg_entry(request):
    key = request.param
    yield spec[key]

@pytest.fixture(params = tuple(SPEC_NOTIMPLEMENTED))
def notimpl_entry(request):
    key  = request.param
    yield spec[key]

def assert_src_array_equal(src, dst):
    if isinstance(src, np.ndarray):
        assert np.array_equal(src, dst)
    elif isinstance(src, pd.DataFrame):
        assert_frame_equal(src, dst)
    elif isinstance(src, pd.Series):
        assert_series_equal(src, dst, check_names = False)
    else:
        assert src == dst
    
# Data ========================================================================
data_dt = data_frame(
    g = ['a', 'a', 'b', 'b'],
    x = pd.to_datetime(["2019-01-01 01:01:01", "2020-04-08 02:02:02", "2021-07-15 03:03:03", "2022-10-22 04:04:04"])
    )

data_str = data_frame(
    g = ['a', 'a', 'b', 'b'],
    x = ['abc', 'cde', 'fg', 'h']
    )

data_bool = data_frame(
    g = ['a', 'a', 'b', 'b'],
    x = [True, False, True, False],
    y = [True, True, False, False]
        )

data_default = data_frame(
    g = ['a', 'a', 'a', 'b', 'b', 'b'],
    x = [10, 11, 11, 13, 13, 13],
    y = [1,2,3,4,5,6]
    )

DATA = data = {
    'dt': data_dt,
    'str': data_str,
    None: data_default,
    'bool': data_bool
}

def get_data(entry, data, backend = None):

    req_bool = entry['result'].get('op') == 'bool'

    # pandas is forgiving to bool inputs
    if isinstance(backend, PandasBackend):
        req_bool = False

    return data['bool'] if req_bool else data[entry['accessor']]


def test_missing_implementation(entry, backend):
    # Check whether test should xfail, skip, or -------------------------------
    backend_status = entry['result'].get(backend.name)

    # case: Needs to be implmented
    # TODO(table): uses xfail
    if backend_status == "xfail":
        pytest.xfail("TODO: impelement this translation")
    
    # case: Can't be used in a mutate (e.g. a SQL ordered set aggregate function)
    # TODO(table): no_mutate
    if backend.name in entry['result'].get('no_mutate', []):
        pytest.skip("Spec'd failure")

    # case: won't be implemented
    if entry['result'].get(backend.name) == "not_impl":
        pytest.skip()


def get_df_expr(entry):
    str_expr = str(entry['expr_frame'])
    call_expr = strip_symbolic(eval(str_expr, {'_': _}))

    return str_expr, call_expr

def cast_result_type(entry, backend, ser):
    sql_type = entry['result'].get('sql_type')
    if isinstance(backend, SqlBackend) and sql_type == 'float':
        return ser.astype('float')
    
    return ser

# Tests =======================================================================


def test_series_against_call(entry):
    if entry['result']['type'] == "Window":
        pytest.skip()

    df = data[entry['accessor']]
    # TODO: once reading from yaml, no need to repr
    str_expr = str(entry['expr_series'])

    call_expr = strip_symbolic(eval(str_expr, {'_': _}))
    res = call_expr(df.x)

    dst = eval(str_expr, {'_': df.x})
    
    assert res.__class__ is dst.__class__
    assert_src_array_equal(res, dst)


def test_frame_expr(entry):
    # TODO: remove this test, not checking anything new
    df = data[entry['accessor']]
    # TODO: once reading from yaml, no need to repr
    str_expr = str(entry['expr_frame'])

    call_expr = strip_symbolic(eval(str_expr, {'_': _}))
    res = call_expr(df)

    dst = eval(str_expr, {'_': df})
    
    assert res.__class__ is dst.__class__
    assert_src_array_equal(res, dst)


def test_pandas_grouped_frame_fast_not_implemented(notimpl_entry):
    from siuba.experimental.pd_groups.dialect import fast_mutate
    gdf = data[notimpl_entry['accessor']].groupby('g')

    # TODO: once reading from yaml, no need to repr
    str_expr = str(notimpl_entry['expr_frame'])
    call_expr = strip_symbolic(eval(str_expr, {'_': _}))

    with pytest.raises(NotImplementedError):
        res = fast_mutate(gdf, result = call_expr)
    


#@backend_pandas
@pytest.mark.skip_backend('sqlite')
def test_frame_mutate(skip_backend, backend, entry):
    test_missing_implementation(entry, backend)

    # Prepare input data ------------------------------------------------------
    # case: inputs must be boolean
    crnt_data = get_data(entry, DATA, backend)
    df = backend.load_df(crnt_data)

    # Execute mutate ----------------------------------------------------------
    str_expr, call_expr = get_df_expr(entry)

    # Run test for equality w/ ungrouped pandas ----
    dst = crnt_data.assign(result = call_expr(crnt_data))
    dst['result'] = cast_result_type(entry, backend, dst['result'])

    assert_equal_query(
            df,
            mutate(result = call_expr),
            dst
            )

    # Run test for equality w/ grouped pandas ----
    g_dst = crnt_data.groupby('g').apply(lambda d: d.assign(result = call_expr)).reset_index(drop = True)
    g_dst['result'] = cast_result_type(entry, backend, g_dst['result'])
    assert_equal_query(
            df,
            group_by(_.g) >> mutate(result = call_expr),
            g_dst
            )


def test_pandas_grouped_frame_fast_mutate(entry):
    from siuba.experimental.pd_groups.dialect import fast_mutate, DataFrameGroupBy
    gdf = get_data(entry, DATA).groupby('g')

    # Execute mutate ----------------------------------------------------------
    str_expr, call_expr = get_df_expr(entry)

    res = fast_mutate(gdf, result = call_expr)
    dst = mutate(gdf, result = call_expr)

    # TODO: apply mark to skip failing tests, rather than downcast
    # pandas grouped aggs, when not using cython, _try_cast back to original type
    # but since mutate uses apply, it doesn't :/. Currently only affects median func.
    dst_obj = dst.obj
    if str_expr == '_.x.median()':
        dst_obj['result'] = gdf._try_cast(dst_obj['result'], gdf.x.obj)

    assert isinstance(dst, DataFrameGroupBy)
    assert_frame_equal(res.obj, dst_obj)


@pytest.mark.skip_backend('sqlite')
def test_frame_summarize(skip_backend, backend, agg_entry):
    entry = agg_entry
    test_missing_implementation(entry, backend)

    # Prepare input data ------------------------------------------------------
    # case: inputs must be boolean
    crnt_data = get_data(entry, DATA, backend)
    df = backend.load_df(crnt_data)

    # Execute mutate ----------------------------------------------------------
    str_expr, call_expr = get_df_expr(entry)

    dst = data_frame(result = call_expr(crnt_data))

    # Process output ----------------------------------------------------------
    # case: output is of a different type than w/ pandas
    dst['result'] = cast_result_type(entry, backend, dst['result'])

    # Run test for equality w/ pandas ----
    # otherwise, verify returns same result as mutate
    assert_equal_query(
            df,
            summarize(result = call_expr),
            dst
            )

    dst_g = crnt_data.groupby('g').apply(call_expr).reset_index().rename(columns = {0: 'result'})
    assert_equal_query(
            df,
            group_by(_.g) >> summarize(result = call_expr),
            dst_g
            )

# Edge Cases ==================================================================

def test_frame_set_aggregates_postgresql():
    # TODO: probably shouldn't be creating backend here
    backend = SqlBackend("postgresql")
    dfs = backend.load_df(data[None])
    
    expr = _.x.quantile(.75)
    assert_equal_query(
            dfs,
            group_by(_.g) >> summarize(result = expr),
            data_frame(g = ['a', 'b'], result = [11., 13.])
            )





from siuba import error
import sys
import pytest


@pytest.fixture(scope = "function")
def last_exc():
    # Note: can't really insert on sys.last_value here, and clean up after...
    err = error.ShortException("testing", short = True)

    yield err

def test_siuba_error_last(last_exc):
    # test part
    assert last_exc.__suppress_context__ is True
    sys.last_value = last_exc
    
    err2 = error.last()

    assert last_exc is err2
    err2.__suppress_context__ is False


def test_siuba_error_shortexception_from_verb():

    err = error.ShortException.from_verb("Mutate", "some_arg_name", "some message", short = True)

    assert "Mutate" in str(err)
    assert "some_arg_name" in str(err)
    assert "some message" in str(err)

    assert err.__suppress_context__ is True

    assert error.ERROR_INSTRUCTIONS in str(err)

def test_siuba_error_instruction(last_exc):
    sys.last_value = last_exc
    full_error_code = str(last_exc).split(":")[-1].lstrip()

    with pytest.raises(error.ShortException) as err:
        exec(full_error_code)



"""
Note: this test file was heavily influenced by its dbplyr counterpart.

https://github.com/tidyverse/dbplyr/blob/master/tests/testthat/test-verb-group_by.R
"""
    
from siuba import (
        _, group_by,
        join, inner_join, left_join, right_join, full_join,
        semi_join, anti_join
        )
from siuba.dply.vector import row_number, n
from siuba.sql.verbs import collect

import pytest
from .helpers import assert_equal_query, assert_frame_sort_equal, data_frame, backend_notimpl, backend_sql


DF1 = data_frame(
        ii = [1,2,3,4],
        x = ["a", "b", "c", "d"]
        )

DF2 = data_frame(
        ii = [1,2,26],
        y = ["a", "b", "z"]
        )

DF3 = data_frame(
        ii = [26],
        z = ["z"]
        )

@pytest.fixture(scope = "module")
def df1(backend):
    return backend.load_df(DF1)

@pytest.fixture(scope = "module")
def df2(backend):
    return backend.load_df(DF2)

@pytest.fixture(scope = "module")
def df2_jj(backend):
    return backend.load_df(DF2.rename(columns = {"ii": "jj"}))

@pytest.fixture(scope = "module")
def df3(backend):
    return backend.load_df(DF3)



@backend_sql("TODO: pandas")
def test_join_diff_vars_keeps_left(backend, df1, df2_jj):
    out = inner_join(df1, df2_jj, {"ii": "jj"}) >> collect()

    assert out.columns.tolist() == ["ii", "x", "y"]

def test_join_on_str_arg(df1, df2):
    out = inner_join(df1, df2, "ii") >> collect()

    target = DF1.iloc[:2,].assign(y = ["a", "b"])
    assert_frame_sort_equal(out, target)

def test_join_on_list_arg(backend):
    # TODO: how to validate how cols are being matched up?
    data = DF1.assign(jj = lambda d: d.ii)
    df_a = backend.load_df(data)
    df_b = backend.load_df(DF2.assign(jj = lambda d: d.ii))
    out = inner_join(df_a, df_b, ["ii", "jj"]) >> collect()

    assert_frame_sort_equal(out, data.iloc[:2, :].assign(y = ["a", "b"]))

@pytest.mark.skip("TODO: note, unsure of this syntax")
def test_join_on_same_col_multiple_times():
    data = data_frame(ii = [1,2,3], jj = [1,2, 9])
    df_a = backend.load_df(data)
    df_b = backend.load_df(data_frame(ii = [1,2,3]))

    out = inner_join(df_a, df_b, {("ii", "jj"): "ii"}) >> collect()
    # keeps all but last row
    assert_frame_sort_equal(out, data.iloc[:2,])

def test_join_on_missing_col(df1, df2):
    with pytest.raises(KeyError):
        inner_join(df1, df2, {"ABCDEF": "ii"})

    with pytest.raises(KeyError):
        inner_join(df1, df2, {"ii": "ABCDEF"})

def test_join_suffixes_dupe_names(df1):
    out = inner_join(df1, df1, {"ii": "ii"}) >> collect()
    non_index_cols = DF1.columns[DF1.columns != "ii"]
    assert all((non_index_cols + "_x").isin(out))
    assert all((non_index_cols + "_y").isin(out))



# Test basic join types -------------------------------------------------------

def test_basic_left_join(df1, df2):
    out = left_join(df1, df2, {"ii": "ii"}) >> collect()
    target = DF1.assign(y = ["a", "b", None, None])
    assert_frame_sort_equal(out, target)

@backend_sql("TODO: pandas returns columns in rev name order")
def test_basic_right_join(backend, df1, df2):
    # same as left join, but flip df arguments
    out = right_join(df2, df1, {"ii": "ii"}) >> collect()
    target = DF1.assign(y = ["a", "b", None, None])
    assert_frame_sort_equal(out, target)

def test_basic_inner_join(df1, df2):
    out = inner_join(df1, df2, {"ii": "ii"}) >> collect()
    target = DF1.iloc[:2,:].assign(y = ["a", "b"])
    assert_frame_sort_equal(out, target)

@pytest.mark.skip_backend("sqlite")
def test_basic_full_join(skip_backend, backend, df1, df2):
    out = full_join(df1, df2, {"ii": "ii"}) >> collect()
    target = DF1.merge(DF2, on = "ii", how = "outer")
    assert_frame_sort_equal(out, target)

def test_basic_semi_join(backend, df1, df2):
    assert_frame_sort_equal(
            semi_join(df1, df2, {"ii": "ii"}) >> collect(),
            DF1.iloc[:2,]
            )

def test_semi_join_no_cross(backend, df1, df2):
    df_ii = backend.load_df(data_frame(ii = [1,1]))
    assert_frame_sort_equal(
            semi_join(df1, df_ii, {"ii": "ii"}) >> collect(),
            DF1.iloc[:1,]
            )

def test_basic_anti_join(backend, df1, df2):
    assert_frame_sort_equal(
            anti_join(df1, df2, on = {"ii": "ii"}) >> collect(),
            DF1.iloc[2:,]
            )

def test_basic_anti_join(backend, df1, df2):
    assert_frame_sort_equal(
            anti_join(df1, df2, on = {"ii": "ii"}) >> collect(),
            DF1.iloc[2:,]
            )

def test_basic_anti_join(backend, df1, df2):
    assert_frame_sort_equal(
            anti_join(df1, df2, on = {"ii": "ii", "x": "y"}) >> collect(),
            DF1.iloc[2:,]
            )



import warnings

from sqlalchemy.sql.elements import ClauseElement 
from sqlalchemy.sql.base import ImmutableColumnCollection
from sqlalchemy import sql

from ..translate import (
        SqlColumn, SqlColumnAgg,
        win_cumul, AggOver, CumlOver, RankOver, warn_arg_default, win_absent
        )
from ..dialects.sqlite import SqliteColumn
from siuba.dply.vector import (
        #cumall, cumany, cummean,
        desc,
        dense_rank, percent_rank, min_rank, cume_dist,
        row_number,
        #ntile,
        between,
        coalesce,
        lead, lag,
        n,
        n_distinct,
        na_if,
        #near,
        nth, first, last
        )


from ..translate import SiubaSqlRuntimeWarning
warnings.simplefilter('once', SiubaSqlRuntimeWarning)

# desc ------------------------------------------------------------------------

@desc.register(ClauseElement)
def _desc_sql(x) -> ClauseElement:
    """
    Example:
        >>> print(desc(sql.column('a')))
        a DESC
    """
    return x.desc()


# ranking functions -----------------------------------------------------------
# note: here we don't use the decorator syntax, but maybe we should for
#       consistency
# TODO: remove repetition in rank definitions

def _sql_rank_over(rank_func, col, partition):
    # partitioning ensures aggregates that use total length are correct
    # e.g. percent rank, cume_dist and friends
    over_clause = RankOver(
            rank_func(),
            order_by = col,
            partition_by = col.isnot(None) if partition else None
            )

    return sql.case({col.isnot(None): over_clause})

def _sql_rank(col, na_option, func_name, partition = False):
    rank_func = getattr(sql.func, func_name)

    if na_option == "keep":
        return _sql_rank_over(rank_func, col, partition = partition)

    warn_arg_default(func_name, 'na_option', None, "keep")

    return RankOver(rank_func(), order_by = col)


@dense_rank.register(ClauseElement)
def _dense_rank_sql(col, na_option = None) -> RankOver:
    return _sql_rank(col, na_option, 'dense_rank')


@percent_rank.register(ClauseElement)
def _percent_rank_sql(col, na_option = None) -> RankOver:
    return _sql_rank(col, na_option, 'percent_rank')
    

@cume_dist.register(ClauseElement)
def _cume_dist_sql(col, na_option = None) -> RankOver:
    return _sql_rank(col, na_option, 'cume_dist', partition = True)


@min_rank.register(ClauseElement)
def _min_rank_sql(col, na_option = None) -> RankOver:
    return _sql_rank(col, na_option, 'rank', partition = True)

dense_rank.register(SqliteColumn, win_absent("DENSE_RANK"))
percent_rank.register(SqliteColumn, win_absent("PERCENT_RANK"))
cume_dist.register(SqliteColumn, win_absent("CUME_DIST"))
min_rank.register(SqliteColumn, win_absent("MIN_RANK"))

# row_number ------------------------------------------------------------------

@row_number.register(ClauseElement)
def _row_number_sql(col) -> CumlOver:
    """
    Example:
        >>> print(row_number(sql.column('a')))
        row_number() OVER ()
        
    """
    return CumlOver(sql.func.row_number())

row_number.register(SqliteColumn, win_absent("ROW_NUMBER"))

# between ---------------------------------------------------------------------

@between.register(ClauseElement)
def _between_sql(x, left, right, default = None) -> ClauseElement:
    """
    Example:
        >>> print(between(sql.column('a'), 1, 2))
        a BETWEEN :a_1 AND :a_2

        >>> print(between(sql.column('a'), 1, 2, default = False))
        coalesce(a BETWEEN :a_1 AND :a_2, :coalesce_1)

    """
    
    if default is not False:
        # TODO: warn
        pass

    if default is None:
        return x.between(left, right)

    return sql.functions.coalesce(x.between(left, right), default)

# coalesce --------------------------------------------------------------------

@coalesce.register(ClauseElement)
def _coalesce_sql(x, *args) -> ClauseElement:
    """
    Example:
        >>> print(coalesce(sql.column('a'), sql.column('b')))
        coalesce(a, b)

        >>> coalesce(1, sql.column('a'))
        Traceback (most recent call last):
            ...
        TypeError: ...
    """
    return sql.functions.coalesce(x, *args)


# lead and lag ----------------------------------------------------------------

@lead.register(ClauseElement)
def _lead_sql(x, n = 1, default = None) -> ClauseElement:
    """
    Example:
        >>> print(lead(sql.column('a'), 2, 99))
        lead(a, :lead_1, :lead_2) OVER (ROWS BETWEEN UNBOUNDED PRECEDING AND CURRENT ROW)
    """
    f = win_cumul("lead")
    return f(x, n, default)

@lag.register(ClauseElement)
def _lag_sql(x, n = 1, default = None) -> ClauseElement:
    """
    Example:
        >>> print(lag(sql.column('a'), 2, 99))
        lag(a, :lag_1, :lag_2) OVER (ROWS BETWEEN UNBOUNDED PRECEDING AND CURRENT ROW)
    """
    f = win_cumul("lag")
    return f(x, n , default)


# n ---------------------------------------------------------------------------

@n.register(ClauseElement)
@n.register(ImmutableColumnCollection)
def _n_sql(x) -> ClauseElement:
    """
    Example:
        >>> print(n(sql.column('a')))
        count(*) OVER ()
    """
    return AggOver(sql.func.count())

@n.register(SqlColumnAgg)
def _n_sql_agg(x) -> ClauseElement:
    """
    Example:
        >>> from siuba.sql.translate import SqlColumnAgg
        >>> print(n(SqlColumnAgg('x')))
        count(*)
    """

    return sql.func.count()


n.register(SqliteColumn, win_absent("N"))
row_number.register(SqliteColumn, win_absent("ROW_NUMBER"))

# n_distinct ------------------------------------------------------------------

@n_distinct.register(ClauseElement)
def _n_distinct_sql(x) -> ClauseElement:
    """
    Example:
        >>> print(n_distinct(sql.column('a')) )
        count(distinct(a))
    """
    return sql.func.count(sql.func.distinct(x))


# na_if -----------------------------------------------------------------------

@na_if.register(ClauseElement)
def _na_if_sql(x, y) -> ClauseElement:
    """
    Example:
        >>> print(na_if(sql.column('x'), 2))
        nullif(x, :nullif_1)
    """
    return sql.func.nullif(x, y)


# nth, first, last ------------------------------------------------------------
# note: first and last wrap around nth, so are not dispatchers.
#       this may need to change this in the future, since this means they won't
#       show their own name, when you print, e.g. first(_.x)

@nth.register(ClauseElement)
def _nth_sql(x, n, order_by = None, default = None) -> ClauseElement:
    if default is not None:
        raise NotImplementedError("default argument not implemented")

    if n < 0 and order_by is not None:
        # e.g. -1 in python is 0, -2 is 1
        n = abs(n + 1)
        order_by = order_by.desc()


    return RankOver(sql.func.nth_value(x, n + 1), order_by = order_by)


@nth.register(SqlColumnAgg)
def _nth_sql_agg(x, n, order_by = None, default = None) -> ClauseElement:
    raise NotImplementedError("nth, first, and last not available in summarize")




from siuba.sql.verbs import collect, show_query, LazyTbl
from siuba.dply.verbs import Pipeable
from .helpers import data_frame
import pandas as pd

import pytest

@pytest.fixture(scope = "module")
def df(backend):
    return backend.load_df(data_frame(x = [1,2,3]))

def test_show_query(df):
    assert isinstance(show_query(df), df.__class__)
    assert isinstance(df >> show_query(), df.__class__)
    assert isinstance(show_query(), Pipeable)

def test_collect(df):
    assert isinstance(collect(df), pd.DataFrame)
    assert isinstance(df >> collect(), pd.DataFrame)
    assert isinstance(collect(), Pipeable)



from siuba.spec.series import spec
from siuba.siu import CallTreeLocal

from siuba.experimental.pd_groups.translate import SeriesGroupBy, GroupByAgg, GROUP_METHODS


# TODO: make into CallTreeLocal factory function

out = {}
for name, entry in spec.items():
    #if entry['result']['type']: continue
        
    key = (entry['result']['type'], entry['data_arity'])
    meth = GROUP_METHODS[key]
    out[name] = meth(
            name = name.split('.')[-1],
            is_property = entry['is_property'],
            accessor = entry['accessor']
            )

call_listener = CallTreeLocal(
        out,
        call_sub_attr = ('str', 'dt', 'cat'),
        chain_sub_attr = True,
        dispatch_cls = GroupByAgg,
        result_cls = SeriesGroupBy
        )


# Fast group by verbs =========================================================

from siuba.siu import Call
from siuba.dply.verbs import mutate, filter, summarize, singledispatch2, DataFrameGroupBy, _regroup

def grouped_eval(__data, expr):
    if isinstance(expr, Call):
        call = call_listener.enter(expr)

        #
        grouped_res = call(__data)
        if isinstance(grouped_res, GroupByAgg):
            res = grouped_res._broadcast_agg_result()
        elif isinstance(grouped_res, SeriesGroupBy):
            res = grouped_res.obj
        else:
            # can happen right now if user selects, e.g., a property of the
            # groupby object, like .dtype, which returns a single value
            # in the future, could restrict set of operations user could perform
            raise ValueError("Result must be subclass of SeriesGroupBy")

    # TODO: for non-call arguments, need to validate they're "literals"
    return res


# Fast mutate ----

@singledispatch2(DataFrameGroupBy)
def fast_mutate(__data, **kwargs):
    """Warning: this function is experimental"""
    out = __data.obj.copy()
    groupings = __data.grouper.groupings

    for name, expr in kwargs.items():
        res = grouped_eval(__data, expr)
        out[name] = res

    return out.groupby(groupings)


@fast_mutate.register(object)
def _fast_mutate_default(__data, **kwargs):
    # TODO: had to register object second, since singledispatch2 sets object dispatch
    #       to be a pipe (e.g. unknown types become a pipe by default)
    # by default dispatch to regular mutate
    f = mutate.registry[type(__data)]
    return f(__data, **kwargs)


# Fast filter ----

@singledispatch2(DataFrameGroupBy)
def fast_filter(__data, *args):
    """Warning: this function is experimental"""
    import pandas as pd
    out = []
    groupings = __data.grouper.groupings

    for expr in args:
        res = grouped_eval(__data, expr)
        out.append(res)

    filter_df = filter.registry[pd.DataFrame]

    df_result = filter_df(__data.obj, *out)

    # TODO: research how to efficiently & robustly subset groupings
    group_names = [ping.name for ping in groupings]
    return df_result.groupby(group_names)


@fast_filter.register(object)
def _fast_filter_default(__data, *args, **kwargs):
    # TODO: had to register object second, since singledispatch2 sets object dispatch
    #       to be a pipe (e.g. unknown types become a pipe by default)
    # by default dispatch to regular mutate
    f = filter.registry[type(__data)]
    return f(__data, *args, **kwargs)


# Fast summarize ----

@singledispatch2(DataFrameGroupBy)
def fast_summarize(__data, **kwargs):
    """Warning: this function is experimental"""
    groupings = __data.grouper.groupings

    # TODO: better way of getting this frame?
    out = __data.grouper.result_index.to_frame()
    
    for name, expr in kwargs.items():
        res = grouped_eval(__data, expr)
        out[name] = res

    out.reset_index(drop = True)
    return out


@fast_summarize.register(object)
def _fast_summarize_default(__data, **kwargs):
    # TODO: had to register object second, since singledispatch2 sets object dispatch
    #       to be a pipe (e.g. unknown types become a pipe by default)
    # by default dispatch to regular mutate
    f = mutate.registry[type(__data)]
    return f(__data, **kwargs)




# sqlvariant, allow defining 3 namespaces to override defaults
from ..translate import (
        SqlColumn, SqlColumnAgg,
        base_scalar, base_agg, base_win, SqlTranslator, 
        win_agg, sql_scalar, sql_agg
        )
import sqlalchemy.sql.sqltypes as sa_types
from sqlalchemy import sql


# Custom dispatching in call trees ============================================

class PostgresqlColumn(SqlColumn): pass

class PostgresqlColumnAgg(SqlColumnAgg, PostgresqlColumn): pass

# Custom translations =========================================================


def sql_log(col, base = None):
    if base is None:
        return sql.func.ln(col)
    return sql.func.log(col)

def sql_round(col, n):
    return sql.func.round(sql.cast(col, sa_types.Numeric()), n)

def sql_func_contains(col, pat, case = True, flags = 0, na = None, regex = True):
    # TODO: warn there differences in regex for python and sql?
    # TODO: validate pat is string?
    if not isinstance(pat, str):
        raise TypeError("pat argument must be a string")
    if flags != 0 or na is not None:
        raise NotImplementedError("flags and na options not supported")

    if not regex:
        case_col = col if case else col.lower()
        return case_col.contains(pat, autoescape = True)

    full_op = "~" if case else "~*"

    return col.op(full_op)(pat)

scalar = SqlTranslator(
        base_scalar,
        log = sql_log,
        round = sql_round,
        contains = sql_func_contains,
        #year = lambda col: sql.func.extract('year', sql.cast(col, sql.sqltypes.Date)),
        concat = sql.func.concat,
        cat = sql.func.concat,
        str_c = sql.func.concat,
        )

aggregate = SqlTranslator(
        base_agg,
        all = sql_agg("bool_and"),
        any = sql_agg("bool_or"),
        std = sql_agg("stddev_samp"),
        var = sql_agg("var_samp"),
        )

window = SqlTranslator(
        base_win,
        any = win_agg("bool_or"),
        all = win_agg("bool_and"),
        lag = win_agg("lag"),
        var = win_agg("var_samp"),
        )

funcs = dict(scalar = scalar, aggregate = aggregate, window = window)



from siuba.siu import _, strip_symbolic, FunctionLookupError, Symbolic, MetaArg
import pytest

def test_op_vars_slice():
    assert strip_symbolic(_.a[_.b:_.c]).op_vars() == {'a', 'b', 'c'}

# Symbolic dispatch ===========================================================
from siuba.siu import symbolic_dispatch, Call, FuncArg

def test_FuncArg():
    f = lambda x: 1
    expr = FuncArg(f)

    assert expr(None) is f

def test_FuncArg_in_call():
    call = Call(
            '__call__',
            FuncArg(lambda x, y: x + y),
            1, y = 2
            )

    assert call(None) == 3



def test_symbolic_dispatch():
    @symbolic_dispatch
    def f(x, y = 2):
        return x + y

    # w/ simple Call
    call1 = f(strip_symbolic(_), 3)
    assert isinstance(call1, Call)
    assert call1(2) == 5

    # w/ simple Symbol
    sym2 = f(_, 3)
    assert isinstance(sym2, Symbolic)
    assert sym2(2) == 5

    # w/ complex Call
    sym3 = f(_['a'], 3)
    assert sym3({'a': 2}) == 5



# Call Tree Local =============================================================
from siuba.siu import CallTreeLocal

@pytest.fixture
def ctl():
    local = {'f_a': lambda self: self}
    yield CallTreeLocal(local, call_sub_attr = ('str', 'dt'))


def test_call_tree_local_sub_attr_method(ctl):
    # sub attr gets stripped w/ method call
    call = ctl.enter(strip_symbolic(_.str.f_a()))
    assert call('x') == 'x'

def test_call_tree_local_sub_attr_property(ctl):
    # sub attr gets stripped w/ property access
    call = ctl.enter(strip_symbolic(_.str.f_a))
    assert call('x') == 'x'

def test_call_tree_local_sub_attr_alone(ctl):
    # attr alone is treated like a normal getattr
    call = ctl.enter(strip_symbolic(_.str))
    assert call.func == "__getattr__"
    assert call.args[1] == "str"

def test_call_tree_local_sub_attr_method_missing(ctl):
    # subattr raises lookup errors (method)
    with pytest.raises(FunctionLookupError):
        ctl.enter(strip_symbolic(_.str.f_b()))

def test_call_tree_local_sub_attr_property_missing(ctl):
    # subattr raises lookup errors (property)
    with pytest.raises(FunctionLookupError):
        ctl.enter(strip_symbolic(_.str.f_b))

# symbolic dispatch and call tree local ----
class SomeClass: pass

@pytest.fixture
def f_dispatch():
    @symbolic_dispatch
    def f(x):
        return 'default'

    @f.register(SomeClass)
    def _f_some_class(x):
        return 'some class'

    yield f


def test_call_tree_local_dispatch_cls_object(f_dispatch):
    ctl = CallTreeLocal(
            {'f_a': lambda self: self},
            dispatch_cls = object
            )

    call = Call("__call__", FuncArg(f_dispatch), MetaArg('_'))
    new_call = ctl.enter(call)
    assert new_call('na') == 'default'


def test_call_tree_local_dispatch_cls_subclass(f_dispatch):
    ctl = CallTreeLocal(
            {'f_a': lambda self: self},
            dispatch_cls = SomeClass
            )

    call = Call("__call__", FuncArg(f_dispatch), MetaArg('_'))
    new_call = ctl.enter(call)
    assert new_call('na') == 'some class'


# strict symbolic dispatch and call tree local ----

@pytest.fixture
def f_dispatch_strict():
    @symbolic_dispatch(cls = SomeClass)
    def f(x):
        return 'some class'

    yield f

def test_strict_dispatch_strict_default_fail(f_dispatch_strict):
    class Other(object): pass

    obj = Other()

    with pytest.raises(TypeError):
        f_dispatch_strict(obj)

def test_call_tree_local_dispatch_fail(f_dispatch_strict):
    ctl = CallTreeLocal(
            {'f_a': lambda self: self},
            dispatch_cls = object
            )

    call = Call("__call__", FuncArg(f_dispatch_strict), MetaArg('_'))

    # should be the default failure dispatch for object
    new_call = ctl.enter(call)
    with pytest.raises(TypeError):
        new_call('na')
 



from siuba import spread, gather, _
from pandas.testing import assert_frame_equal
import pandas as pd

import pytest

@pytest.fixture
def df():
    return pd.DataFrame({
        'id': [1, 1, 2],
        'm': ['a', 'b', 'b'],
        'v': [1,2,3]
        })

@pytest.fixture
def wide_df():
    return pd.DataFrame({
        'id': [1,2],
        'a': [1, None],
        'b': [2., 3.]
        })


@pytest.mark.parametrize('key, value', [
    ("m", "v"),
    (_.m, _.v),
    (_["m"], _["v"]),
    (1, 2),
    ])
def test_spread_selection(df, wide_df, key, value):
    res = spread(df, key, value)


    assert_frame_equal(res, wide_df)
    

def test_spread_fill(df, wide_df):
    res = spread(df, "m", "v", fill = 99)

    assert_frame_equal(res, wide_df.fillna(99.))


def test_spread_grouped_df(df, wide_df):
    gdf = df.groupby('id')
    res = spread(gdf, "m", "v")

    gdf_wide = wide_df.groupby('id')
    assert_frame_equal(res.obj, gdf_wide.obj)

    assert len(res.grouper.groupings) == 1
    assert res.grouper.groupings[0].name == "id"


def test_gather(df, wide_df):
    res = gather(wide_df, "m", "v", "a", "b") 

    # note, unlike df, includes an NA now
    long = pd.DataFrame({
        'id': [1,2,1,2],
        'm': ['a', 'a', 'b', 'b'],
        'v': [1, None, 2, 3]
        })

    assert_frame_equal(res, long)


def test_gather_drop_na(df, wide_df):
    res = gather(wide_df, "m", "v", "a", "b", drop_na = True) 

    # note that .dropna doesn't work here, since coerces floats to ints
    assert_frame_equal(res, df.assign(v = df.v.astype(float)))


def test_gather_drop_na_varselect(df, wide_df):
    res = gather(wide_df, "m", "v", _["a": "b"], drop_na = True) 

    # note that .dropna doesn't work here, since coerces floats to ints
    assert_frame_equal(res, df.assign(v = df.v.astype(float)))




from .verbs import *

# preceed w/ underscore so it isn't exported by default
# we just want to register the singledispatch funcs
from .dply import vector as _vector
from .dply import string as _string



from setuptools import setup, find_packages

# parse version ---------------------------------------------------------------

import re
import ast
_version_re = re.compile(r'__version__\s+=\s+(.*)')

with open('siuba/__init__.py', 'rb') as f:
    version = str(ast.literal_eval(_version_re.search(
        f.read().decode('utf-8')).group(1)))

# setup -----------------------------------------------------------------------

setup(
    name='siuba',
    packages=find_packages(),
    version=version,
    description='A package for quick, scrappy analyses with pandas and SQL',
    author='Michael Chow',
    license='MIT',
    author_email='mc_al_gh_siuba@fastmail.com',
    url='https://github.com/machow/siuba',
    keywords=['package', ],
    install_requires=[
        "pandas>=0.24.0",
        "numpy>=1.12.0",
        "SQLAlchemy>=1.1.18",
        "psycopg2>=2.7.0",
    ],
    python_requires=">=3.6",
    include_package_data=True,
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)




import pandas as pd
import numpy as np
from siuba.siu import symbolic_dispatch

# fct_reorder -----------------------------------------------------------------

@symbolic_dispatch
def fct_reorder(fct, x, func = np.median, desc = False) -> pd.Categorical:
    """Return copy of fct, with categories reordered according to values in x.
    
    Arguments:
        fct: a pandas.Categorical, or array(-like) used to create one.
        x: values used to reorder categorical. Must be same length as fct.
        func: function run over all values within a level of the categorical.
        desc: whether to sort in descending order.

    Examples:
        >>> fct_reorder(['a', 'a', 'b'], [4, 3, 2])
        [a, a, b]
        Categories (2, object): [b, a]

        >>> fct_reorder(['a', 'a', 'b'], [4, 3, 2], desc = True)
        [a, a, b]
        Categories (2, object): [a, b]

        >>> fct_reorder(['x', 'x', 'y'], [4, 0, 2], np.max)
        [x, x, y]
        Categories (2, object): [y, x]

    """

    x_vals = x.values if isinstance(x, pd.Series) else x
    s = pd.Series(x_vals, index = fct)

    # for each cat, calc agg func, make values of ordered the codes
    ordered = s.groupby(level = 0).agg(func).sort_values(ascending = not desc)
    ordered[:] = np.arange(len(ordered))
    codes = ordered[s.index.values]
    return pd.Categorical.from_codes(codes, list(ordered.index))


# fct_recode ------------------------------------------------------------------

@symbolic_dispatch
def fct_recode(fct, recat=None, **kwargs) -> pd.Categorical:
    """Return copy of fct with renamed categories.

    Arguments:
        fct: a pandas.Categorical, or array(-like) used to create one. 
        **kwargs: arguments of form new_name = old_name.

    Examples:
        >>> cat = ['a', 'b', 'c']
        >>> fct_recode(cat, z = 'c')
        [a, b, z]
        Categories (3, object): [a, b, z]

        # >>> fct_recode(cat, x = 'a', x = 'b')
        # >>> fct_recode(cat, x = ['a', 'b'])
        
    """

    if recat and not isinstance(recat, dict):
        raise TypeError("fct_recode requires named args or a dict.")

    if recat and kwargs:
        duplicate_keys = set(recat).intersection(set(kwargs))
        if duplicate_keys:
            raise ValueError(
                "The following recode name(s) were specified more than once: {}" \
                .format(duplicate_keys)
            )

    new_cats = {**recat, **kwargs} if recat else kwargs
    return fct_collapse(fct, new_cats)


# fct_collapse ----------------------------------------------------------------

@symbolic_dispatch
def fct_collapse(fct, recat, group_other = None) -> pd.Categorical:
    """Return copy of fct with categories renamed. Optionally group all others.

    Arguments:
        fct: a pandas.Categorical, or array(-like) used to create one.  
        recat: dictionary of form {new_cat_name: old_cat_name}. old_cat_name may be
               a list of existing categories, to be given the same name.
        group_other: an optional string, specifying what all other categories should be named.

    Notes:
        Resulting levels index is ordered according to the earliest level replaced.
        If we rename the first and last levels to "c", then "c" is the first level.

    Examples:
        >>> fct_collapse(['a', 'b', 'c'], {'x': 'a'})
        [x, b, c]
        Categories (3, object): [x, b, c]

        >>> fct_collapse(['a', 'b', 'c'], {'x': 'a'}, group_other = 'others')
        [x, others, others]
        Categories (2, object): [x, others]

        >>> fct_collapse(['a', 'b', 'c'], {'ab': ['a', 'b']})
        [ab, ab, c]
        Categories (2, object): [ab, c]

    """
    if not isinstance(fct, pd.Categorical):
        fct = pd.Categorical(fct)

    # each existing cat will map to a new one ----
    # need to know existing to new cat
    # need to know new cat to new code
    cat_to_new = {k: None for k in fct.categories}
    for new_name, v in recat.items():
        v = [v] if not np.ndim(v) else v
        for old_name in v:
            if cat_to_new[old_name] is not None:
                raise Exception("category %s was already re-assigned"%old_name)
            cat_to_new[old_name] = new_name

    # collapse all unspecified cats to group_other if specified ----
    for k, v in cat_to_new.items():
        if v is None:
            if group_other is not None:
                cat_to_new[k] = group_other
            else:
                cat_to_new[k] = k

    # map from old cat to new code ----
    # calculate new codes
    ordered_cats = {new: True for old, new in cat_to_new.items()}
    new_cat_set = {k: ii for ii, k in enumerate(ordered_cats)}
    # map old cats to new codes
    remap_code = {old: new_cat_set[new] for old, new in cat_to_new.items()}

    new_codes = fct.map(remap_code)
    new_cats = list(new_cat_set.keys())
    return pd.Categorical.from_codes(new_codes, new_cats)


# fct_lump --------------------------------------------------------------------

@symbolic_dispatch
def fct_lump(fct, n = None, prop = None, w = None, other_level = "Other", ties = None) -> pd.Categorical:
    """
    Arguments:
        fct: a pandas.Categorical, or array(-like) used to create one.
        n: number of categories to keep.
        prop: (not implemented) keep categories that occur prop proportion of the time.
        w: array of weights corresponding to each value in fct.
        other_level: name for all lumped together levels.
        ties: (not implemented) method to use in the case of ties.

    Notes:
        Currently, one of n and prop must be specified.

    Examples:
        >>> fct_lump(['a', 'a', 'b', 'c'], n = 1)
        [a, a, Other, Other]
        Categories (2, object): [a, Other]

        # TODO: implement prop arg
        #>>> fct_lump(['a', 'a', 'b', 'b', 'c', 'd'], prop = .2)
        #[a, a, b, b, Other, Other]
        #Categories (3, object): [b, a, Other]
        

    """

    if ties is not None:
        raise NotImplementedError("ties is not implemented")

    if n is None and prop is None:
        raise NotImplementedError("Either n or prop must be specified")

    if prop is not None:
        raise NotImplementedError("prop arg is not implemented")

    keep_cats = _fct_lump_n_cats(fct, n, w, other_level, ties)
    return fct_collapse(fct, {k:k for k in keep_cats}, group_other = other_level)

def _fct_lump_n_cats(fct, n, w, other_level, ties):
    # TODO: currently always selects n, even if ties
    ascending = n < 0
    arr = _get_values(w) if w is not None else 1
    ser = pd.Series(arr, index = fct)
    sorted_arr = ser.groupby(level = 0).sum().sort_values(ascending = ascending)
    return sorted_arr.iloc[:abs(n)].index.values

def _get_values(x):
    # TODO: move into utility, note pandas now encouraging .array method
    if isinstance(x, pd.Series): return x.values

    return x


# fct_rev ---------------------------------------------------------------------

@symbolic_dispatch
def fct_rev(fct) -> pd.Categorical:
    """Return a copy of fct with category level order reversed.next
    
    Arguments:
        fct: a pandas.Categorical, or array(-like) used to create one.

    """

    if not isinstance(fct, pd.Categorical):
        fct = pd.Categorical(fct)

    rev_levels = list(reversed(fct.categories))

    return fct.reorder_categories(rev_levels)
